@@ -1,7 +1,7 @@
 /*
  * regexp.c:
  *
- * Copyright (C) 2007-2011 David Lutterkort
+ * Copyright (C) 2007-2015 David Lutterkort
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -50,7 +50,7 @@ char *regexp_escape(const struct regexp *r) {
     ret = fa_restrict_alphabet(r->pattern->str, strlen(r->pattern->str),
                                &nre, &nre_len, 2, 1);
     if (ret == 0) {
-        pat = escape(nre, nre_len);
+        pat = escape(nre, nre_len, RX_ESCAPES);
         free(nre);
     }
 #endif
@@ -237,10 +237,11 @@ regexp_union(struct info *info, struct regexp *r1, struct regexp *r2) {
 }
 
 char *regexp_expand_nocase(struct regexp *r) {
-    const char *p = r->pattern->str;
+    const char *p = r->pattern->str, *t;
     char *s = NULL;
     size_t len;
     int ret;
+    int psub = 0, rsub = 0;
 
     if (! r->nocase)
         return strdup(p);
@@ -248,16 +249,55 @@ char *regexp_expand_nocase(struct regexp *r) {
     ret = fa_expand_nocase(p, strlen(p), &s, &len);
     ERR_NOMEM(ret == REG_ESPACE, r->info);
     BUG_ON(ret != REG_NOERROR, r->info, NULL);
+
+    /* Make sure that r->pattern->str and ret have the same number
+     * of parentheses/groups, since our parser critically depends
+     * on the fact that the regexp for a union/concat and those
+     * of its children have groups that are in direct relation */
+    for (t = p; *t; t++) if (*t == '(') psub += 1;
+    for (t = s; *t; t++) if (*t == '(') rsub += 1;
+    BUG_ON(psub < rsub, r->info, NULL);
+    psub -= rsub;
+    if (psub > 0) {
+        char *adjusted = NULL, *a;
+        if (ALLOC_N(adjusted, strlen(s) + 2*psub + 1) < 0)
+            ERR_NOMEM(true, r->info);
+        a = adjusted;
+        for (int i=0; i < psub; i++) *a++ = '(';
+        a = stpcpy(a, s);
+        for (int i=0; i < psub; i++) *a++ = ')';
+        free(s);
+        s = adjusted;
+    }
  error:
     return s;
 }
 
+static char *append_expanded(struct regexp *r, char **pat, char *p,
+                             size_t *len) {
+    char *expanded = NULL;
+    size_t ofs = p - *pat;
+    int ret;
+
+    expanded = regexp_expand_nocase(r);
+    ERR_BAIL(r->info);
+
+    *len += strlen(expanded) - strlen(r->pattern->str);
+
+    ret = REALLOC_N(*pat, *len);
+    ERR_NOMEM(ret < 0, r->info);
+
+    p = stpcpy(*pat + ofs, expanded);
+ error:
+    FREE(expanded);
+    return p;
+}
+
 struct regexp *
 regexp_union_n(struct info *info, int n, struct regexp **r) {
     size_t len = 0;
     char *pat = NULL, *p, *expanded = NULL;
     int nnocase = 0, npresent = 0;
-    int ret;
 
     for (int i=0; i < n; i++)
         if (r[i] != NULL) {
@@ -284,14 +324,8 @@ regexp_union_n(struct info *info, int n, struct regexp **r) {
             *p++ = '|';
         *p++ = '(';
         if (mixedcase && r[i]->nocase) {
-            expanded = regexp_expand_nocase(r[i]);
+            p = append_expanded(r[i], &pat, p, &len);
             ERR_BAIL(r[i]->info);
-            len += strlen(expanded) - strlen(r[i]->pattern->str);
-            ret = REALLOC_N(pat, len);
-            ERR_NOMEM(ret < 0, info);
-            p = pat + strlen(pat);
-            p = stpcpy(p, expanded);
-            FREE(expanded);
         } else {
             p = stpcpy(p, r[i]->pattern->str);
         }
@@ -320,7 +354,6 @@ regexp_concat_n(struct info *info, int n, struct regexp **r) {
     size_t len = 0;
     char *pat = NULL, *p, *expanded = NULL;
     int nnocase = 0, npresent = 0;
-    int ret;
 
     for (int i=0; i < n; i++)
         if (r[i] != NULL) {
@@ -345,14 +378,8 @@ regexp_concat_n(struct info *info, int n, struct regexp **r) {
             continue;
         *p++ = '(';
         if (mixedcase && r[i]->nocase) {
-            expanded = regexp_expand_nocase(r[i]);
+            p = append_expanded(r[i], &pat, p, &len);
             ERR_BAIL(r[i]->info);
-            len += strlen(expanded) - strlen(r[i]->pattern->str);
-            ret = REALLOC_N(pat, len);
-            ERR_NOMEM(ret < 0, info);
-            p = pat + strlen(pat);
-            p = stpcpy(p, expanded);
-            FREE(expanded);
         } else {
             p = stpcpy(p, r[i]->pattern->str);
         }
