@@ -1,7 +1,7 @@
 /*
  * test-api.c: test public API functions for conformance
  *
- * Copyright (C) 2009-2011 Red Hat Inc.
+ * Copyright (C) 2009-2015 David Lutterkort
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -121,6 +121,13 @@ static void testGet(CuTest *tc) {
     CuAssertPtrNotNull(tc, value);
     CuAssertIntEquals(tc, AUG_NOERROR, aug_error(aug));
 
+    /* aug_get should set VALUE to NULL even if the path expression is invalid
+       Issue #372 */
+    value = (const char *) 7;
+    r = aug_get(aug, "[invalid path]", &value);
+    CuAssertIntEquals(tc, -1, r);
+    CuAssertPtrEquals(tc, NULL, value);
+
     aug_close(aug);
 }
 
@@ -642,6 +649,20 @@ static void testAugEscape(CuTest *tc) {
     free(out);
 }
 
+static void testRm(CuTest *tc) {
+    struct augeas *aug;
+    int r;
+
+    aug = aug_init(root, loadpath, AUG_NO_STDINC|AUG_NO_MODL_AUTOLOAD);
+    CuAssertPtrNotNull(tc, aug);
+
+    r = aug_set(aug, "/files/1/2/3/4/5", "1");
+    CuAssertRetSuccess(tc, r);
+
+    r = aug_rm(aug, "/files//*");
+    CuAssertIntEquals(tc, 5, r);
+}
+
 int main(void) {
     char *output = NULL;
     CuSuite* suite = CuSuiteNew();
@@ -661,6 +682,7 @@ int main(void) {
     SUITE_ADD_TEST(suite, testTextStore);
     SUITE_ADD_TEST(suite, testTextRetrieve);
     SUITE_ADD_TEST(suite, testAugEscape);
+    SUITE_ADD_TEST(suite, testRm);
 
     abs_top_srcdir = getenv("abs_top_srcdir");
     if (abs_top_srcdir == NULL)
